# -*- coding: utf-8 -*-
#
# w.c.s. - web application for online forms
# Copyright (C) 2005-2013  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import re
import xml.etree.ElementTree as ET
import zipfile

from .misc import date_format, datetime_format
from .evalutils import make_date, make_datetime

FO_NS = 'urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0'
OFFICE_NS = 'urn:oasis:names:tc:opendocument:xmlns:office:1.0'
STYLE_NS = 'urn:oasis:names:tc:opendocument:xmlns:style:1.0'
NUMBER_NS = 'urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0'
TABLE_NS = 'urn:oasis:names:tc:opendocument:xmlns:table:1.0'
TEXT_NS = 'urn:oasis:names:tc:opendocument:xmlns:text:1.0'
XLINK_NS = 'http://www.w3.org/1999/xlink'


def is_number(value):
    try:
        float(value)
    except ValueError:
        return False
    return True


class Workbook(object):
    def __init__(self, encoding='utf-8'):
        self.sheets = []
        self.encoding = encoding

    def add_sheet(self, name):
        sheet = WorkSheet(self, name)
        self.sheets.append(sheet)
        return sheet

    def get_content_node(self):
        root = ET.Element('{%s}document-content' % OFFICE_NS)
        ET.SubElement(root, '{%s}scripts' % OFFICE_NS)
        ET.SubElement(root, '{%s}font-face-decls' % OFFICE_NS)

        body = ET.SubElement(root, '{%s}body' % OFFICE_NS)
        spreadsheet = ET.SubElement(body, '{%s}spreadsheet' % OFFICE_NS)
        for sheet in self.sheets:
            spreadsheet.append(sheet.get_node())
        return root

    def get_styles_node(self):
        root = ET.Element('{%s}document-styles' % OFFICE_NS)
        ET.SubElement(root, '{%s}font-face-decls' % OFFICE_NS)
        automatic_styles = ET.SubElement(root, '{%s}styles' % OFFICE_NS)
        style = ET.SubElement(automatic_styles, '{%s}style' % STYLE_NS)
        style.attrib['{%s}name' % STYLE_NS] = 'Default'

        def define_date_style(name, strftime_string):
            node = ET.SubElement(automatic_styles, '{%s}date-style' % NUMBER_NS)
            node.attrib['{%s}name' % STYLE_NS] = name + 'NumberFormat'
            for part in re.findall(r'%?.', strftime_string):
                if part == '%Y':
                    ET.SubElement(node, '{%s}year' % NUMBER_NS)
                elif part == '%m':
                    ET.SubElement(node, '{%s}month' % NUMBER_NS)
                elif part == '%d':
                    ET.SubElement(node, '{%s}day' % NUMBER_NS)
                elif part == '%H':
                    ET.SubElement(node, '{%s}hours' % NUMBER_NS)
                elif part == '%M':
                    ET.SubElement(node, '{%s}minutes' % NUMBER_NS)
                elif part == '%S':
                    ET.SubElement(node, '{%s}seconds' % NUMBER_NS)
                else:
                    ET.SubElement(node, '{%s}text' % NUMBER_NS).text = part
            style = ET.SubElement(automatic_styles, '{%s}style' % STYLE_NS)
            style.attrib['{%s}name' % STYLE_NS] = name
            style.attrib['{%s}family' % STYLE_NS] = 'table-cell'
            style.attrib['{%s}data-style-name' % STYLE_NS] = name + 'NumberFormat'
            style.attrib['{%s}parent-style' % STYLE_NS] = 'Default'

            style = ET.SubElement(automatic_styles, '{%s}style' % STYLE_NS)
            style.attrib['{%s}name' % STYLE_NS] = name + 'Column'
            style.attrib['{%s}family' % STYLE_NS] = 'table-column'
            ET.SubElement(style, '{%s}table-column-properties' % STYLE_NS).attrib['{%s}column-width' % STYLE_NS] = '80mm'

        define_date_style('Date', date_format())
        define_date_style('DateTime', datetime_format())

        return root

    def get_styles(self):
        return ET.tostring(self.get_styles_node(), 'utf-8')

    def get_content(self):
        return ET.tostring(self.get_content_node(), 'utf-8')

    def save(self, output):
        z = zipfile.ZipFile(output, 'w')
        z.writestr('content.xml', self.get_content())
        z.writestr('styles.xml', self.get_styles())
        z.writestr('mimetype', 'application/vnd.oasis.opendocument.spreadsheet')
        z.writestr('META-INF/manifest.xml', '''<?xml version="1.0" encoding="UTF-8"?>
<manifest:manifest xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0">
 <manifest:file-entry manifest:full-path="/" manifest:media-type="application/vnd.oasis.opendocument.spreadsheet"/>
 <manifest:file-entry manifest:full-path="styles.xml" manifest:media-type="text/xml"/>
 <manifest:file-entry manifest:full-path="content.xml" manifest:media-type="text/xml"/>
 <manifest:file-entry manifest:full-path="META-INF/manifest.xml" manifest:media-type="text/xml"/>
 <manifest:file-entry manifest:full-path="mimetype" manifest:media-type="text/plain"/>
</manifest:manifest>''')
        z.close()


class WorkSheet(object):
    def __init__(self, workbook, name):
        self.cells = {}
        self.name = name
        self.workbook = workbook

    def write(self, row, column, value, **kwargs):
        if not row in self.cells:
            self.cells[row] = {}
        self.cells[row][column] = WorkCell(self, value, **kwargs)

    def get_node(self):
        root = ET.Element('{%s}table' % TABLE_NS)
        root.attrib['{%s}name' % TABLE_NS] = self.name
        ET.SubElement(root, '{%s}table-column' % TABLE_NS)
        for i in range(0, max(self.cells.keys())+1):
            row = ET.SubElement(root, '{%s}table-row' % TABLE_NS)
            for j in range(0, max(self.cells.get(i).keys())+1):
                cell = self.cells.get(i, {}).get(j, None)
                if not cell:
                    ET.SubElement(row, '{%s}table-cell' % TABLE_NS)
                else:
                    row.append(cell.get_node())
        return root


class WorkCell(object):
    def __init__(self, worksheet, value, formdata=None, data_field=None, native_value=None):
        if value is None:
            value = ''
        if type(value) is not unicode:
            value = unicode(value, 'utf-8')
        self.worksheet = worksheet
        self.value = value
        self.formdata = formdata
        self.data_field = data_field
        self.native_value = native_value

    def get_node(self):
        root = ET.Element('{%s}table-cell' % TABLE_NS)
        root.attrib['{%s}value-type' % OFFICE_NS] = 'string'
        p = ET.SubElement(root, '{%s}p' % TEXT_NS)
        p.text = self.value
        if not self.data_field:
            return root

        if not self.native_value:
            return root

        if self.data_field.type == 'file':
            url = '%sfiles/%s/%s' % (
                    self.formdata.get_url(backoffice=True),
                    self.data_field.id, self.native_value.base_filename)
            a = ET.SubElement(p, '{%s}a' % TEXT_NS)
            a.attrib['{%s}href' % XLINK_NS] = url
            a.text = self.native_value.base_filename
            p.text = None
        elif self.data_field.type in ('time', 'last_update_time'):
            root.attrib['{%s}style-name' % TABLE_NS] = 'DateTime'
            root.attrib['{%s}value-type' % OFFICE_NS] = 'date'
            root.attrib['{%s}date-value' % OFFICE_NS] = make_datetime(self.native_value).strftime('%Y-%m-%dT%H:%M:%S')
            p.text = make_datetime(self.native_value).strftime(datetime_format())
        elif self.data_field.type == 'date':
            root.attrib['{%s}style-name' % TABLE_NS] = 'Date'
            root.attrib['{%s}value-type' % OFFICE_NS] = 'date'
            root.attrib['{%s}date-value' % OFFICE_NS] = make_date(self.native_value).strftime('%Y-%m-%d')
            p.text = make_datetime(self.native_value).strftime(date_format())
        elif is_number(self.value):
            root.attrib['{%s}value-type' % OFFICE_NS] = 'float'
            root.attrib['{%s}value' % OFFICE_NS] = str(self.value)

        return root
